/* -*-C++-*- */
/*
 *  Copyright (c) 1999 Regents of the University of California.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgment:
 *       This product includes software developed by the Imaging and
 *       Distributed Collaboratories Group at Lawrence Berkeley
 *       National Laboratory.
 *  4. Neither the name of the University nor of the Laboratory may be used
 *     to endorse or promote products derived from this software without
 *     specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 *
 */

#include <iostream>
#include <string>
#include "udpping.h"
#include "Results.h"
#include "PingResults.h"
#include "timefuncs.h"

using std::cout;
using std::cerr;
using std::endl;
using std::string;

#define MAXPINGMSG     1024


/* ---------------------------------------------------------
 *  UDPPing:
 *	serv - boolean indicator of whether this is the test server
 *      address - address to send the pings to and expect them from
 *      port - port to use for the ping test
 *      delay - msec delay between pings
 *
 * ---------------------------------------------------------*/
  
UDPPing::UDPPing( PingTestInfo testinf ): MyDebug("UDPPing"), mPinginf(testinf)
{
    // how many msecs to delay between each ping test
    // minimum delay will be the round trip time
    mDelaypings = static_cast<int>(1000/mPinginf.getSendRate());  

    for( int i=0; i < AVGWINDOW; i++)
      mLastrtt[i] = 0.0;

    mNumbertests = 0;  // how many ping tests have been run
    mTestlimit   = static_cast<int>( mPinginf.getTestAmount() );   // number of ping tests
    mUdpcon      = UDPNet( 0, mDelaypings, mPinginf.getPort() );  // init the ping socket

    printDebug(3, "Port=%d  Local Name = %s\n", mPinginf.getPort(), (mPinginf.getLocalName()).c_str() );
}

UDPPing::~UDPPing()
{
}


/* ---------------------------------------------------------
 *  sendPingTest: 
 *	Conducts server side of ping test
 *	repetitive ping test
 *
 * --------------------------------------------------------- */

int UDPPing::sendPingTest()
{
    struct timeval lasttime, curtime;  // time most recent ping was sent and received respectively
    char buf[MAXPINGMSG];  // buffer for storing incoming ping message
    int err;
    int cursleeptime;  // measured in msecs, time to sleep before sending next ping
    int temputime;     // temp variable used in calculating time to sleep between pings
    double prevrtt = -1.0;  // last ping round trip time
    int numberfail = 0;  // number of consecutive pings that have not returned

    string sRemote( mPinginf.getRemoteTestPeer() );

    printDebug(3, "Remote Host= %s\n", sRemote.c_str());
    mUdpcon.openSocket( sRemote );  // open the ping socket
    mUdpcon.setBlockTime( mDelaypings/1000, ((mDelaypings%1000)*1000) );  // set the receive blocking time

    mNumbertests = 0;
    //sleep( mPinginf.getDelay() );
    printDebug(3, "about to start sendPingTest %d\n", mTestlimit);

    string format("%s %s %ld %ld %f");

    // either send pings forever or upto the limit set
    while( 1 ){
	// send ping test
	gettimeofday( &lasttime, NULL );

	sprintf( buf, format.c_str(), (mPinginf.getLocalName()).c_str(), 
		 sRemote.c_str(), lasttime.tv_sec, lasttime.tv_usec, prevrtt );

	err = mUdpcon.doSend( buf, strlen( buf ));
	printDebug(3, "Sent message: %s\n", buf );

	if( mNumbertests == mTestlimit )
	    break;
	
	err = mUdpcon.doRecv( buf, MAXPINGMSG );
	gettimeofday( &curtime, NULL );

	if( err > 0 ){
	    printDebug(3, "received ping- %s\n", buf );
	    prevrtt = calculateRTT( buf, &curtime );
            printDebug(3, "%d) %d) %f \n", mNumbertests, mNumbertests%AVGWINDOW, prevrtt);
	    numberfail = 0;
	}
	else
	    numberfail++;

	// if the client has not bounced back a ping in 20 seconds
	if( numberfail > (20000 / mDelaypings )){
	    cerr << "Failed to receive a ping from the client in " << 20000/mDelaypings  << " closing connection to client" 
		 << endl;
	    break;
	}
	
	cursleeptime = mDelaypings - (curtime.tv_sec - lasttime.tv_sec )*1000;
	temputime = (curtime.tv_usec - lasttime.tv_usec )/1000;
	if( temputime > 0 )
	    cursleeptime -= temputime;
	
	if( mNumbertests != 0 && cursleeptime > 0 )
	    usleep( cursleeptime * 1000 );
    }

    return mNumbertests;
}

/* ---------------------------------------------------------
 *  bouncePingTest - conducts client side of ping test
 *  This routine receives pings in and bounces them back to the
 *  originator.  Also in the received ping packet is the most
 *  recent ping results.  
 * ---------------------------------------------------------*/

int UDPPing::bouncePingTest()
{
    struct timeval rcvtime;  // time the ping packet came in
    struct timeval sndtime;  // time when the ping packet is about to be sent
    struct timeval difftime;  // time spent processing the ping packet
    char buf[MAXPINGMSG];    // the ping packet is stored here
    char buf2[MAXPINGMSG];   // used to compose difftime to add to ping packet
    int err;
    char *send_sec, *send_usec; // what time the ping packet was sent by the server
    char *serv, *receiver;  // ping packet from and to fields
    char *msg;       // temporary buffer pointer used while parsing ping packet
    char *prev;      // last ping round trip time
    double dblprev;  // last ping round trip time converted to a double
    int   numberfail = 0;  // number of consecutive lost ping packets

    string bRemote( mPinginf.getRemoteTestPeer() );

    err = mUdpcon.openSocket( bRemote );  // open socket for ping test
    if( err < 0 ){
	cerr << "mUdpcon.openSocket( " << bRemote << " ) failed \n";
	return -1;
    }
    printDebug(3, "bouncePingTest socket open\n");

    // wait a minimum of 1 second for the ping packet
    if( mDelaypings < 1000 )
        mUdpcon.setBlockTime( 1, 0 );
    else
        mUdpcon.setBlockTime( mDelaypings/1000, ((mDelaypings%1000)*1000) );  // set the receive blocking time
    mNumbertests = 0;  // loop counter

    string format1("%s %s %s %s");
    string format2(" %ld %ld");

    printDebug(3, "about to start bounce tests %d\n", mTestlimit);
    // either send pings forever or upto the limit set
    while( mTestlimit == -1 || mNumbertests < mTestlimit ){
	// receive ping test	
	err = mUdpcon.doRecv( buf, MAXPINGMSG );
	
	// keep track of failures and quit if have too many
	if( err <= 0 ){
	    cerr << "UDPPING:: doRecv failed with err = " << err << endl;
	    numberfail++;
	    if( numberfail > 10 ){
		cerr<< "Failing to receive pings for 10 consecutive attempts from the server, quiting\n";
	        break;
	    }
	    continue;
	}

	numberfail = 0;

	// record the ping packet receive time
	gettimeofday( &rcvtime, NULL );

	printDebug(3, "Message received is: %s\n", buf );

	// check that sender and receiver are correct
	msg = (char *)calloc( sizeof(buf), sizeof(char) );
        strncpy( msg, buf, sizeof(buf) );

	serv = strtok( msg, " " );
 	receiver = strtok( NULL, " " );
 	send_sec = strtok( NULL, " " );
 	send_usec = strtok( NULL, " " );
 	prev = strtok( NULL, " " );

	// check whether this ping is from my server and intended for me
	if( strcmp(bRemote.c_str(), serv ))
	    printDebug(3, "This does not appear to be from my server %s %s\n", bRemote.c_str(), serv);
	if(  strcmp( (mPinginf.getLocalName()).c_str(), receiver ))
	    printDebug(3,"This does not appear to be for me %s\n", (mPinginf.getLocalName()).c_str());

	// convert the previous ping round trip time to a double and record it
	sscanf( prev, "%lf", &dblprev );
	if( dblprev >= 0.0 ){
	    //pthread_mutex_lock( &mStats_mutex );
	    printDebug(3, "%d) %d) %f\n", mNumbertests, 
				mNumbertests%AVGWINDOW, dblprev);
	    mLastrtt[mNumbertests%AVGWINDOW] = dblprev;
	    mNumbertests++;
	    //pthread_mutex_unlock( &mStats_mutex );
	}

	// fill in the ping packet
	sprintf( buf, format1.c_str(), serv, receiver, send_sec, send_usec);

	// need to determine how much time we have been processing this ping packet
	gettimeofday( &sndtime, NULL );
	difftime = subtimev( &sndtime, &rcvtime );

	// put the time we have spent processing the ping packet into the packet
	// so the other end can subtract it off of the ping round trip time.
	sprintf( buf2, format2.c_str(), difftime.tv_sec, difftime.tv_usec );
	strcat( buf, buf2 );

	// send the ping back to the server
	mUdpcon.doSend( buf, strlen( buf ));
	printDebug(3, "Message returned is: %s\n", buf );

	free( msg );
    }

    return mNumbertests;
}

/* ---------------------------------------------------------
 *  calculateRTT - called by the ping server side to calculate
 *  the round trip time after a ping packet has bounced back.
 *  arguments
 *        pingmsg - the ping packet as it was received
 *        rcvtime - the time when this packet was received
 * ---------------------------------------------------------*/

double 
UDPPing::calculateRTT( char *pingmsg, struct timeval *rcvtime)
{
    int    index = 0;
    struct timeval sndtime;
    struct timeval delay;
    char   sender[MAXHOSTNAMELEN];
    char   receiver[MAXHOSTNAMELEN];
    struct timeval rtt;

    sscanf( pingmsg, "%s %s %ld %ld %ld %ld", sender, receiver,
	    &(sndtime.tv_sec), &(sndtime.tv_usec), &(delay.tv_sec),
	    &(delay.tv_usec) );

    rtt = subtimev( &(subtimev( rcvtime, &sndtime )), &delay );

    index = mNumbertests%AVGWINDOW;
    //pthread_mutex_lock( &mStats_mutex );
    mLastrtt[index] = timevtomsec( &rtt );
    mNumbertests++;
    //pthread_mutex_unlock( &mStats_mutex );

    return mLastrtt[index];
}

/* ---------------------------------------------------------
 *  ThreadMain - this is the routine that is called when the
 *  thread is run using pthread_create.  It launches the ping
 *  test by calling the appropriate routine.
 * ---------------------------------------------------------*/
 
void 
UDPPing::ThreadMain( void )
{
    int status;

    printDebug(3, "Starting ping test\n");
    // start up the ping test

    if( mPinginf.isSender() )
	status = sendPingTest( );
    else
        status = bouncePingTest( );
    if( status < 0 )
      {
	  cerr << "ERROR running udpping" << endl;
	  return;
      }

    printDebug(3, "Done with ping test\n");
    return;
}


/* ---------------------------------------------------------
 *  printStats - this routine prints the current test
 *  results.  It can be called asynchronously and it is
 *  also called when the ping test finishes.  This routine
 *  is thread safe (a mutex locks the statistics when they
 *  are being updated)
 * ---------------------------------------------------------*/

double 
UDPPing::printStats( void )
{
    int    index = 0;
    int i = 0;
    double maxrtt, minrtt, avgrtt, curlast;
    
    index = (mNumbertests%AVGWINDOW) - 1;

    //pthread_mutex_lock( &mStats_mutex );
    avgrtt = 0.0;
    for( i=0; i< AVGWINDOW && i < mNumbertests ; i++ ){
	if( i == 0 ){
	    maxrtt = mLastrtt[i];
	    minrtt = mLastrtt[i];
	}
	else{
	    if( maxrtt < mLastrtt[i] )
	        maxrtt = mLastrtt[i];
	    if( minrtt > mLastrtt[i] )
	        minrtt = mLastrtt[i];
	}
	avgrtt += mLastrtt[i];
    }
    curlast = mLastrtt[index];

    // remove the high value and the low value from the average
    if( i > 4 ){
	// remove the high value and the low value from the average
	avgrtt = avgrtt - minrtt - maxrtt;
	i -= 2;
    }
    avgrtt = avgrtt / ((double)i);

    cout << "Unicast RTT (msec) to " << mPinginf.getRemoteTestPeer() << endl;
    cout << " is: avg= "<< avgrtt  << endl;
    cout << " max= "    << maxrtt  << endl;
    cout << " min= "    << minrtt  << endl;
    cout << " latest= " << curlast << endl;

    return curlast;
}


/* ---------------------------------------------------------
 * double UDPPing::getAverageRTT()
 *
 * ---------------------------------------------------------*/

double 
UDPPing::getAverageRTT( void )
{
    double avgrtt = 0.0;
    int i;

    for( i=0; i< AVGWINDOW && i < mNumbertests ; i++ )
    	avgrtt += mLastrtt[i];
    avgrtt = avgrtt / ((double)i);
    return avgrtt;
}


/* ---------------------------------------------------------
 *  printResults
 *
 * ---------------------------------------------------------*/

void
UDPPing::printResults( void )
{
    PingResults pingResults ( mPinginf );
    pingResults.fillStats( mLastrtt, mNumbertests );
    //(mPinginf.getResults())->newPingResults( &mPinginf, mStats_mutex, mLastrtt, mNumbertests );
    (mPinginf.getResults())->newPingResults( pingResults );
}


/* ---------------------------------------------------------
 *  printVariables
 *
 * ---------------------------------------------------------*/

void
UDPPing::printVariables( void )
{
    cout << "***UDPPing:: printVariables\n";
    cout << "Local Host=" << mPinginf.getLocalName() << endl; // the address of this machine
    cout << "Remote   =" << mPinginf.getRemoteTestPeer() << endl; // address to send the pings to (may be multicast)
    cout << "mDelaypings=" << mDelaypings << endl; // how many msecs to delay between each ping test (
                        // minimum delay will be the round trip time
    cout << "Server?    =" << mPinginf.isSender() << endl;     // boolean indicating whether this is the sender for the test
    cout << "mNumbertests=" << mNumbertests << endl;  // how many ping tests have been run so far
    cout << "mTestlimit  =" << mTestlimit << endl;    // number of ping tests to run
    cout << "PingTestInfo mPinginf\n";
    mPinginf.printVariables( cout );   // the properties for this ping test
    cout << "***\n";
}
